# here we set up the keywords allowed (understood) in the config file.
# we have three kinds of keywords: those that we expect once in a 'record',
# (singleton), and those that can appear multiple times (array),
# and finally we have the special 'include' keyword.
# Effectively, this gives us single values for singleton keywords, and
# an array (list) of values for array keywords.
# Below we say for each 'group' which fields we expect, of what kind.

proc config_init {} {
	global config
	config_config_include include include
	config_config_singleton     campaign comment dir makefile config execdir specdir primerdir impldir mkinclude
	config_config_array     campaign experiment
	config_config_singleton experiment runs seed maxdepth pre post driver primer adapter combinator partitioner impl log msg dir config makefile mutant mkinclude
	config_config_array     experiment var driverparams preparams postparams
	config_config_singleton driver configgen exec execdir pre post
	config_config_array     driver genparams execparams preparams postparams
	config_config_singleton spec file language dialect
	config_config_array     spec auxfile
	config_config_singleton primer spec gen exec execdir
	config_config_array     primer channel genparams execparams
	config_config_singleton combinator config gen exec execdir
	config_config_array     combinator primer partitioner tp test combinator genparams execparams
	config_config_singleton partitioner config gen primer partfile exec execdir
	config_config_array     partitioner execparams partitioner
	config_config_singleton impl pre post gen exec execdir execcontext configgen
	config_config_array     impl iap genparams execparams execcontextparams configgenparams preparams postparams
	config_config_singleton adapter codingdir exec execdir
	config_config_array     adapter pco execparams
	config_config_singleton address value name
	config_config_singleton channel iokind timeout
	config_config_array channel pco
	config_config_singleton pco address encoder decoder multiplexer regexp ievent oevent
	config_config_singleton iap address
	config_config_singleton var name type
	config_config_array var value
	config_config_singleton foreach name begin step end
	config_config_array foreach type value
	config_config_singleton product type prefix template
	config_config_array product foreach
}
proc config_finish {} {
	em_compile
}
proc config_cleanup {} {
	catch {unset config}
}

proc config_config_kind {key kind args} {
	global config
	foreach a $args {
		if {![info exists config(fields,$key,$kind,$a)]} {
			lappend config(fields,$key,$kind) $a
			set config(fields,$key,$kind,$a) 1
		}
	}
}

proc config_config_singleton {key args} {
#	eval config_config_kind $key singleton $key
	eval config_config_kind $key singleton base
	eval config_config_kind $key singleton $args
}

proc config_config_array {key args} {
	eval config_config_kind $key array $args
}

proc config_config_include {key args} {
	eval config_config_kind $key include $args
}

proc config_has_children {key} {
	global config
	foreach kind {singleton array include} {
		if {[info exists config(fields,$key,$kind)] &&
		    [llength $config(fields,$key,$kind)] > 0} {
			return 1
		}
	}
	return 0
}

proc config_from_file {filename} {
	if {[catch {open $filename r} fileId]} {
		warningMsg "Cannot open $filename: $fileId"
	} else {
		config_process_file $fileId $filename
		close $fileId
	}
}

proc config_process_file {channel filename} {
	foreach line [split [read $channel] \n] {
		config_process_line $line $filename
	}
	config_handle_block $filename
}

proc config_mkfilename {filename incfile} {
	set t [file pathtype $incfile]
	switch -- $t {
		absolute {
			return $incfile
		}
		relative {
			set dir [file dirname $filename]
			return [file join $dir $incfile]
		}
		default {
			warningMsg "config_mkfilename: unexpected path type: $t for: $incfile"
		}
	}
}

proc config_process_line {line filename} {
	global configtmp
	if {[regexp {^\s*$} $line]} {         # empty line
		return
	}
	if {[regexp {^\s*#} $line]} {         # comment line
		return
	}
	if {[regexp {^[^ \t]} $line]} {       # start of new, end of prev block
		config_handle_block $filename
	}
	if {![info exists configtmp]} {
		set configtmp {}
	}
	set configtmp [concat $configtmp [fields_split $line " \t"]]
}

proc config_handle_block {filename} {
	global configtmp
	global config
	if {![info exists configtmp]} {
		return
	}
	debugMsg config "Fields: $configtmp"
	set key_value [field_first $configtmp]
	if {[llength $key_value] == 2} {
		set key [lindex $key_value 0]
		set value [lindex $key_value 1]
		if {![info exists config(field,present,list,$key)]} {
			lappend config(field,list) $key
			set config(field,present,list,$key) 1
		}
		if {![info exists config(field,present,list,$key,$value)]} {
			lappend config(field,list,$key) $value
			set config(field,present,list,$key,$value) 1
		} else {
			warningMsg "ignoring duplicate block $key=$value"
			catch {unset configtmp}
			return
		}
		if {![info exists config(field,list,$key,$value)]} {
			set config(field,list,$key,$value) {}
		}
		# see if we are dealing with an include statement
		if {[info exists config(fields,$key,include)]} {
			foreach name $config(fields,$key,include) {
				set v [field_value $configtmp $name]
				if {[string compare $v ""] != 0} {
					debugMsg config "  $name: $v"
				}
				set incfile [config_mkfilename $filename $v]
				debugMsg config "Reading $incfile ..."
				catch {unset configtmp}
				config_from_file $incfile
				debugMsg config "... returning to $filename"
			}
			# should we return here?
		}
		# handle singleton fields
		if {[info exists config(fields,$key,singleton)]} {
			foreach name $config(fields,$key,singleton) {
				set v [field_value $configtmp $name]
				if {[string compare $v ""] != 0} {
					debugMsg config "  $name: $v"
					if {![info exists config(field,present,list,$key,$value,$name)]} {
						lappend config(field,list,$key,$value) $name
						set config(field,present,list,$key,$value,$name) 1
					}
					set config(field,list,$key,$value,$name) $v
				}
			}
		}
		# handle array fields
		if {[info exists config(fields,$key,array)]} {
			foreach name $config(fields,$key,array) {
				set v [field_value_array $configtmp $name]
				set l [llength $v]
				if {($l> 0) &&
				    (($l > 1) || ([string compare [lindex $v 0] ""] != 0)) } {
					debugMsg config "  $name: $v"
					if {![info exists config(field,present,list,$key,$value,$name)]} {
						lappend config(field,list,$key,$value) $name
						set config(field,present,list,$key,$value,$name) 1
					}
					set config(field,list,$key,$value,$name) $v
				}
			}
		}
	}
	catch {unset configtmp}
}
