# =========================================================================
#
# simple command interpreter
#

#
# The utcmdline_interpret routine should return an event, in the same
# way as is done by uttrace_read_event.
# As input it gets an event in the usual form - if an event was read from
# a trace, or the empty list - if no event was read from a trace.

# It may set maininfo(stopnow) to stop the whole thing, i.e. to quit
# It probably should keep its own administration if it wants to do
# an x number of steps 'automatically'.

# What we want to do is the following:
# unless we work in 'automatic mode'
#  - print a prompt - 
#  - wait for commands, and execute them, until we get a command that
#    implies a 'next', by which time the user may or may not have chosen
#    an event - if so, return it, if not, return empty list
#    (and let the random selection do its work)

proc utcmdline_interpret {tr_event} {
	global maininfo cmdinterpinfo tester
	set result {} ;# the empty event
	set go_on 1
	set command ""
	set mode normal     ;# possible values: normal or auto
	set auto 0          ;# default: in normal mode
	set autotrace 0     ;# default
	# Warning: we reference $cmdinterpinfo(timeout,$mode) so
	#          be sure to set the timeouts for all possible modes!!
	set cmdinterpinfo(timeout,normal) -1 ;# disable timeout
	set cmdinterpinfo(timeout,auto) 0
	while {$go_on} {
		if {($cmdinterpinfo(mode,auto)) &&
		    (($cmdinterpinfo(stepstodo) > 0) ||
		     ($cmdinterpinfo(stepstodo) == -1))} {
			set auto 1
			set mode auto
			if {$cmdinterpinfo(automode,usetrace)} {
				set autotrace 1
			}
		}
		if {([info exists commandline]) &&
		    ([string compare $commandline ""] != 0)} {
			debugLvl 3 "continuing: <$commandline>\n"
		} else {
			if {! $auto} {
				send_user $cmdinterpinfo(prompt)
				if {$maininfo(gui)} {
					send_user "\n"
					set tester(output,initialnewline) 0
				} else {
					set tester(output,initialnewline) 1
				}
			}
			set timeout $cmdinterpinfo(timeout,$mode)
			# exp_internal 1
			set commandline ""
			# as soon as we have 'observed' eof on stdin via expect_user,
			# we can no longer write to stdout via send_user.
			# so, in batch mode, where we still want to write
			# to stdout, but stdin is closed from the start,
			# we must avoid testing for eof on expect_user.
			# in interactive mode we need expect_user
			# to allow the user to give commands, and
			# and for stopping an 'auto' command prematurely.
			# we assume we are in batch mode when
			#     $maininfo(nrofstepsmax) != 0 && $auto
			if {($maininfo(nrofstepsmax) == 0 || ! $auto) && !$tester(usereof)} {
				expect_user {
					-re "^(\[^\r\n]*)\r*\n" {
						set tester(output,initialnewline) 0
						set commandline [string trim $expect_out(1,string)]
					}
					timeout {
					}
					eof {
						if {$maininfo(nrofstepsmax) == 0 || ! $auto} {
							set maininfo(stopnow) 1
							catch {send_user "\n"}
							set go_on 0
						} else {
							set tester(usereof) 1
						}
					}
				}
			}
			# exp_internal 0
			set timeout $maininfo(timeout)
		}
		set commandargs [split $commandline " "]
		catch {unset commandline}
		set command [lindex $commandargs 0]
		set args [string trim [join [lrange $commandargs 1 end] " "]]
		switch -- $command {
		  stop {
		    set cmdinterpinfo(mode,auto) 0
		    set cmdinterpinfo(automode,usetrace) 0
		    set auto 0
		    set autotrace 0
		    set mode normal
		  }
		  quit -
		  exit {
		    set maininfo(stopnow) 1
			set go_on 0
		  }
		  menusize {
		    set inputs [lindex [spec_get_inputs] 1]
		    set outputs [lindex [spec_get_outputs] 1]
		    set s ""
		    append s "nr-of-inputs: [llength $inputs]\n"
		    append s "nr-of-outputs: [llength $outputs]\n"
		    utlog_touser_output $s
		  }
		  menu {
		    set inputs [lindex [spec_get_inputs] 1]
		    utlog_touser_menu input $inputs
		    set outputs [lindex [spec_get_outputs] 1]
		    utlog_touser_menu output $outputs
		  }
		  path {
		    set trace [utlog_trace]
		    utlog_touser_output [join $trace "\n"]
		  }
		  trace {
		    set tr_io_kind [uttrace_iokind $tr_event]
		    if {[string compare $tr_io_kind ""] != 0} {
		    	set side [lindex $tr_event 0]
		    	set tr_c [lindex $tr_event 1]
		    	set tr_ev [lindex $tr_event 2]
		    	set tr_susp [lindex $tr_event 3]
				set repr [utlog_tostring_event_tuple $side $tr_c $tr_ev $tr_susp ""]
				utlog_touser_output "trace: [set side][set repr]"
			} else {
				utlog_touser_error "(sorry, not tracing, so no trace event)"
			}
		  }
		  usetrace {
		  	# use the trace event
		    set tr_io_kind [uttrace_iokind $tr_event]
		    if {[string compare $tr_io_kind ""] != 0} {
		    	set result $tr_event
		    	set go_on 0
		    } else {
				utlog_touser_error "(sorry, not tracing, so no trace event to use)"
		    }
		  }
		  io {
		    set iokind [spec_get_iokind ""]
		    utlog_touser_output "io: $iokind"
			set result [list $iokind {}]
		  }
		  step -
		  next {
		    if {[regexp -- "^(\[^ \t\]+)(\[ \t\]+(.*))?" $args match subcommand subwhite subargs]} {
		    	# send_user "subcommand: <$match> <$subcommand> <$subwhite> <$subargs>\n"
				if {[regexp -- "^\[0-9\]+$" $subcommand]} {
				    set commandline "auto $subcommand"
				} elseif {[string compare $subcommand "trace"] == 0} {
				    set commandline "usetrace"
				} elseif {[string compare $subcommand "input"] == 0} {
				    set commandline "input $subargs"
				} elseif {[string compare $subcommand "output"] == 0} {
				    set commandline "output"
				} elseif {[string compare $subcommand "event"] == 0} {
					set result [list input $subargs]
					set go_on 0
				} else {
					set go_on 0
				}
			} else {
				set go_on 0
			}
		  }
		  input {
		    set inputs [lindex [spec_get_inputs] 1]
		    if {[llength $inputs] > 0} {
		    	if {[string compare $args ""] != 0} {
		    		if {[regexp "^\\((\[^ \t\\(\\)\]+)\\):?\$" [lindex [split $args] 0] m ch]} {
		    			set chan $ch
		    			set args [join [lrange [split $args] 1 end] " "]
		    		} else {
		    			set chan [default_channel input]
		    		}
			} else {
		    		set chan [default_channel input]
		    		set args {}
		    	}
			if {[regexp {^([^[]*)[ \t]+\[([^]]*)\]}  $args match ev preds]} {
				debugLvl 1 "split ($args) -> ($ev) ($preds)"
				# pattern above should catch space at end of ev;
				# just to be sure we also trim it (trim can go?)
				set entry [list $chan [string trim $ev] 0 $preds]
				set idx [lsearch $inputs $entry]
				debugLvl 1 "lsearch idx $idx entry($entry) inputs($inputs)"
				if {$idx >= 0} {
					set args $ev
				}
			}
			set result [list input $chan $args ]
		    	set go_on 0
		    } else {
		    	utlog_touser_error "(sorry, input menu is empty)"
		    }
		  }
		  output {
		    set outputs [lindex [spec_get_outputs] 1]
		    if {[llength $outputs] > 0} {
		    	set result [list output {} ]
		    	set go_on 0
		    } else {
		    	utlog_touser_error "(sorry, output menu is empty)"
		    }
		  }
		  statesize {
		    set sz 0
			set state [spec_get_state]
			switch -- [lindex $state 0] {
			  state {
			    # subtract two lines from the size,
			    # one for marquage
			    # one for contexte
				set sz [expr [llength [lindex $state 1]] - 2]
			  } error {
				utlog_touser_error "error from specification"
			  }
		    }
		    utlog_touser_output "state-size: $sz"
		  } 
		  state {
			set state [spec_get_state]
			switch -- [lindex $state 0] {
			  state {
				utlog_touser_output [join [lindex $state 1] "\n"]
			  } error {
				utlog_touser_error "error from specification"
			  }
		    }
		  } 
		  debug {
		    # invoke expect debugger right here right now
			debug $args
		  }
		  tcl {
		    ## Tcl command is superflous, because the debugger offers the same
#		    ## and more...
#		    # execute arbitrary tcl code
#		    if {[catch {eval $args} foo]} {
#		    	utlog_touser_error "Tcl: $foo"
#		    } else {
#		    	utlog_touser_error "$foo"
#		    }
#		  }
		  batch {
		    if {[string compare [lindex $commandargs 1] ""] != 0} {
		      if {[regexp -- "^\[0-9\]+$" [lindex $commandargs 1]]} {
		      	set nr [lindex $commandargs 1]
			batch_loop $nr
		      } else {
		      	warning "argument of 'batch' should be a number\n"
		      }
		     } else {
		      	warning "'batch' needs number argument\n"
		    }
		  }
		  auto {
		    set cmdinterpinfo(mode,auto) 1
		    set cmdinterpinfo(automode,usetrace) 0
		    if {[string compare [lindex $commandargs 1] ""] != 0} {
		      if {[regexp -- "^\[0-9\]+$" [lindex $commandargs 1]]} {
		      	set cmdinterpinfo(stepstodo) [lindex $commandargs 1]
		      } else {
		      	warning "argument of 'auto' should be a number\n"
		      }
		    } else {
		      set cmdinterpinfo(stepstodo) -1
		    }
		    if {$cmdinterpinfo(stepstodo) > 0} {
		    	incr cmdinterpinfo(stepstodo) -1 ;# for next (current) step
		    	set result {}  ;# don't trace, even when present
		    	set go_on 0
		    } elseif {$cmdinterpinfo(stepstodo) == -1} {
		    	set result  {}  ;# don't trace, even when present
		    	set go_on 0
		    } else {
		    	warning "cannot auto, because nr-of-steps == 0\n"
		    }
		  }
		  autotrace {
		    set cmdinterpinfo(mode,auto) 1
		    set cmdinterpinfo(automode,usetrace) 1
		    if {[string compare [lindex $commandargs 1] ""] != 0} {
		      if {[regexp -- "^\[0-9\]+$" [lindex $commandargs 1]]} {
		      	set cmdinterpinfo(stepstodo) [lindex $commandargs 1]
		      } else {
		      	warning "argument of 'auto' should be a number\n"
		      }
		    } else {
		      set cmdinterpinfo(stepstodo) -1
		    }
		    set tr_io_kind [uttrace_iokind $tr_event]
		    if {[string compare $tr_io_kind ""] != 0} {
		    	set tmpresult $tr_event
			    if {$cmdinterpinfo(stepstodo) > 0} {
			    	incr cmdinterpinfo(stepstodo) -1 ;# for next (current) step
			    	set result $tmpresult  ;# use trace when present
			    	set go_on 0
			    } elseif {$cmdinterpinfo(stepstodo) == -1} {
			    	set result $tmpresult  ;# use trace when present
			    	set go_on 0
			    } else {
			    	warning "cannot autotrace, because nr-of-steps == 0\n"
			    }
		    } else {
			    	utlog_touser_error "(sorry, not tracing, so no trace event to use)"
		    }
		  }
		  help {
		    utlog_touser_output "recognized commands:
menusize             - print the size of the unsorted menu
statesize            - print the size of the state(vector) (internal format)
menu                 - print the menu (distinquishes inputs from outputs)
state                - print the state(vector) (internal format)
path                 - print the path (the events done from start till now)
trace                - print the current event from the trace that we follow
usetrace             - use the current trace event for next step
next input-event     - do one step (using input-event, when in menu)
next                 - do one step (using the result of last 'io' command)
step input-event     - synonym for 'next input-event'
step                 - synonym for 'next'
io                   - select randomly input or output, to be used in next step
input (chan) event   - do one input step (using event, when in menu)
input event          - do one input step (using event, when in menu)
input (chan)         - do one input step, from channel chan
input                - do one input step, from randomly chosen channel
output (chan)        - do one output step, from channel chan
output               - do one output step, from randomly chosen channel
auto                 - do steps, randomly choosing in- and outputs
                       until end-of-test, or until interrupted by the 'stop'
                       command
auto nr-of-steps     - as auto, but do at most nr-of-steps steps
autotrace            - do steps, following the trace,
                       until end-of-test, or end-of-trace,
                       or until interrupted by the 'stop' command
autotrace nr-steps   - as autotrace, but do at most nr-steps steps
stop                 - interrupt the 'auto' or 'autotrace' command
quit                 - (clean up and) exit the tester
exit                 - synonym for 'quit'
help                 - print this list of commands
"
		  }
		  "" {
		    if {$auto} {
			    if {$cmdinterpinfo(stepstodo) > 0} {
			  		incr cmdinterpinfo(stepstodo) -1
			    } elseif {$cmdinterpinfo(stepstodo) == -1} {
			    	# nothing to do
			    }
			    if {$autotrace} {
			    	# use trace when available
			    	set tr_io_kind [uttrace_iokind $tr_event]
			    	if {[string compare $tr_io_kind ""] != 0} {
			    		set result $tr_event
			    		set go_on 0
			    	} else {
		    			set cmdinterpinfo(mode,auto) 0
		    			set cmdinterpinfo(automode,usetrace) 0
						set auto 0
						set autotrace 0
						set mode normal
			    		# send_user "(end of trace reached)\n"
					}
				} else {
			    	set go_on 0
			    }
		    }
		  }
		  default {
		    utlog_touser_error "Error: unknown command: $command"
		  }
		}
	}
	utlog_tofile_mode $mode
	return $result
}
