/*
 * Traces  grammar, Lex specification
 *
 *  ------------------------------------------------------------------------
 */
 
D			[0-9]
L			[a-zA-Z_]
H			[a-fA-F0-9]
E			[Ee][+-]?{D}+
FS			(f|F|l|L)
IS			(u|U|l|L)*

%{
#include <stdio.h>
#include <string.h>

#include "parser.h"
#include "buf.h"

static Buf *txtline;

static void count(void);
static void describe(void);
static void comment(void);
static void check_type(void);
static casestring stripmkcasestring(char *s);

%}

%%
"/*"				{ comment(); }

^"%description"		{ describe(); return(K_DESCRIPTION);}
^"%declare"			{ count(); return(K_DECLARE);}
^"%define"			{ count(); return(K_EXPRESSION);}
^"%objective"		{ count(); return(K_OBJECTIVE);}
"action"			{ count(); return(K_ACTION);}
"set"				{ count(); return(K_SET);}
".."				{ count(); return(K_TO); }
"infinity"			{ count(); return(K_INFINITY); }

{D}+				{ count();
					  yylval.yt_int=atoi(yytext);
					  return(NUMBER);
					}   
{L}({L}|{D})*		{ count(); check_type(); return(IDENTIFIER); }


\"(\\.|[^\\"])*\"	{ count();
					  yylval.yt_casestring=stripmkcasestring(yytext);
					  return(STRING_LITERAL);
					}

[ \t\v\n\f\r]		{ count(); }
.					{ count(); return yytext[0]; }

%%

int
yywrap()
{
	return(1);
}


int column = 0;
int lineno = 1;

static casestring
stripmkcasestring(char *s)
{
	int l;
	casestring res;
	char c;

	l = strlen(s);
	c = s[l-1];
	s[l-1] = '\0';
	res = mkcasestring(&s[1]);
	s[l-1] = c;
	return res;
}

static void
comment(void)
{
	int c, c1;

	column += strlen(yytext);
	bufcat(txtline,yytext);
	while ((c = input()) != EOF)
		switch ((char) c) {
		case '\n':
			lineno++;
			column = 0;
			bufcpy(txtline, "");
			break;
		case '\t':
			column += 8 - (column % 8);
			bufaddc(txtline,c);
			break;
		case '*':
			column++;
			bufaddc(txtline,c);
			if ((c1=input()) == '/') {
				column++;
				bufaddc(txtline,c1);
				return; 
			} else if (c1 == EOF) {
				warning("unexpected end of file in comment");
				return;
			} else
				unput(c1);
			break;
		default:
			column++;
			bufaddc(txtline,c);
			break;	       
		}
	if (c == EOF)
		warning("unexpected end of file in comment");
}

static void
describe(void)
{
	int c;

	column += strlen(yytext);
	bufcat(txtline,yytext);
	while ((c = input()) != EOF) 
		switch ((char) c) {
		case '\n':
			lineno++;
			column = 0;
			bufcpy(txtline, "");
			break;
		case '\t':
			column += 8 - (column % 8);
			bufaddc(txtline,c);
			break;
		case '%' :
			if (column == 0) {
				unput(c);
				unput('\n');
				lineno--;
				return;
			} else {
				bufaddc(txtline,c);
				column++;
			}
			break;	
		default:
			column++;
			bufaddc(txtline,c);
			break;	       
		}
	if (c == EOF)
		warning("unexpected end of file in %description");
}

static void
count(void)
{
	char *p;

	for (p = yytext; *p != '\0'; p++)
		switch (*p) {
		case '\n':
			lineno++;
			column = 0;
			bufcpy(txtline, "");
			break;
		case '\t':
			column += 8 - (column % 8);
			bufaddc(txtline,*p);
			break;
		default:
			column++;
			bufaddc(txtline,*p);
			break;
		}
}

static void
check_type(void)
{
	yylval.yt_id = Id(Uid(mkcasestring(yytext)));
	yylval.yt_id->column = column;
	yylval.yt_id->lineno = lineno;
}

void
lexerinit()
{
	txtline = bufnew();
}

char*
lexerline()
{
	return bufstring(txtline);
}
