#include <stdlib.h>
#include <stdio.h>
#include "spexRange.h"

RANGE_VALUE_TYPE rangeShiftValue;

// Private function prototypes
//int RCmp(DRangeP *r1, DRangeP *r2);
int RCmp(const void *v1, const void *v2) {
	DRangeP r1 = *(DRangeP *)v1;
	DRangeP r2 = *(DRangeP *)v2;
	return RangeCmp(r1, r2);
}


// Create a range element with lower- and upper-bound.
DRangeP	RangeMakeSingle(RANGE_VALUE_TYPE value_lb, RANGE_VALUE_TYPE value_ub) {
	DRangeP elem;
	
	if (value_lb > value_ub) {
		printf("Error. Tried to create a range with lb (%d) greater \
than upper bound (%d).\n", value_lb, value_ub);
		exit(EXIT_FAILURE);
	}
	
	elem = malloc(sizeof(DRange));
	if (elem == NULL) {
		printf("Error. Could not allocate memory for DRange.\n");
		exit(EXIT_FAILURE);
	}
	elem->lb = value_lb;
	elem->ub = value_ub;
	
	return elem;
}

// Create an array of ranges, from an array of values. The array of values must 
// form 'nrPairs' pairs of two values. 
// (I.e. the array must consist of (2 * nrPairs) values.)
// The array returned will be sorted.
DRangeP *	RangeMakeArray(RANGE_VALUE_TYPE values[], int nrPairs) {
	DRangeP *elem;
	int i;
	
	elem = calloc(nrPairs, sizeof(DRangeP));
	if (elem == NULL) {
		printf("Error. Could not allocate memory for %d DElements.\n", nrPairs);
		exit(EXIT_FAILURE);
	}
	for (i = 0; i < nrPairs; i ++) {
		elem[i] = RangeMakeSingle(values[2*i], values[(2*i)+1]);
	}
	
	// Sort the array
	qsort(elem, i, sizeof(DRangeP), RCmp);
	
	return elem;
}


char	RangeContains(DRangeP r1, RANGE_VALUE_TYPE value) {
	return ( ((r1->lb <= value) && (value <= r1->ub)) );
}

// TODO Possible to inline (#define) this function
char	RangeEq(DRangeP r1, DRangeP r2) {
	return ((r1->lb == r2->lb) && (r1->ub == r2->ub));
}
	
// Overlap means either the lb of r2 is in the range r1, or the ub of r2 is in the range r1
char	RangeHasOverlap(DRangeP r1, DRangeP r2) {
	return ( ((r1->lb <= r2->lb) && (r2->lb <= r1->ub)) || \
	         ((r1->lb <= r2->ub) && (r2->ub <= r1->ub)) || \
	         ((r2->lb <= r1->lb) && (r1->ub <= r2->ub))    );
}

// 'Can combine' means that the ranges are adjoining or overlapping
char	RangeCanCombine(DRangeP r1, DRangeP r2) {
	return ( ((r1->lb - 1 <= r2->lb) && (r2->lb <= r1->ub + 1)) || \
	         ((r1->lb - 1 <= r2->ub) && (r2->ub <= r1->ub + 1)) );
		
}

// Checks if r2 is contained (completely) in r1
char	RangeContainedIn(DRangeP r1, DRangeP r2) {
	return ( (r1->lb <= r2->lb) && (r2->ub <= r1->ub));
}

// Compare ranges. Will return 0 if the ranges are equal, negative if r1 is smaller than r2, or
// positive otherwise.
// If the lb of r1 and r2 is equal, the upper bounds are compared. Otherwise, the lowerbounds
// are compared.
int	RangeCmp(DRangeP r1, DRangeP r2) {
	if (r1->lb == r2->lb) {
		return (r1->ub - r2->ub);
	} else {
		return (r1->lb - r2->lb);
	}
}

// Will return true if r1 is smaller than r2 => (r1->ub < r2->lb)
int	RangeSmallerThan(DRangeP r1, DRangeP r2) {
	return (r1->ub < r2->lb);
}

// Will return true if r1 is larger than r2 => (r1->lb > r2->ub)
int	RangeLargerThan(DRangeP r1, DRangeP r2) {
	return (r1->lb > r2->ub);
}


// Delete all values in r2 from r1. Will return an (allocated) array with two DRangeP entries.
// The first range will contain all values from r1 that were smaller than the lowerbound 
// of r2, or NULL if there were no such values. The second range will contain all values from
// r1 that were larger than the upperbound of r2, or NULL of there were no such values.
DRangeP * RangeDelete(DRangeP r1, DRangeP r2) {
	DRangeP *retRanges;
	
	retRanges = calloc(2, sizeof(DRangeP));
	
	if ( r1->lb < r2->lb ) {
		retRanges[0] = RangeMakeSingle(r1->lb, r2->lb - 1);
	}
	
	if ( r1->ub > r2->ub ) {
		retRanges[1] = RangeMakeSingle(r2->ub + 1, r1->ub);
	}
	
	return retRanges;
}


DRangeP	RangeOverlap(DRangeP r1, DRangeP r2) {
	RANGE_VALUE_TYPE new_lb;
	RANGE_VALUE_TYPE new_ub;
	
	if (! RangeHasOverlap(r1, r2) ) {
		return NULL;
	} else {
		// new_lb = max(r1->lb, r2->lb);
		new_lb = (r1->lb > r2->lb) ? r1->lb : r2->lb;
		// new_lb = min(r1->ub, r2->ub)
		new_ub = (r1->ub < r2->ub) ? r1->ub : r2->ub;
		return RangeMakeSingle(new_lb, new_ub);
	}
}
	
DRangeP RangeCombine(DRangeP r1, DRangeP r2) {
	RANGE_VALUE_TYPE new_lb;
	RANGE_VALUE_TYPE new_ub;
	
	if (! RangeCanCombine(r1, r2) ) {
		return NULL;
	} else {
		// new_lb = min(r1->lb, r2->lb);
		new_lb = (r1->lb < r2->lb) ? r1->lb : r2->lb;
		// new_lb = max(r1->ub, r2->ub)
		new_ub = (r1->ub > r2->ub) ? r1->ub : r2->ub;
		return RangeMakeSingle(new_lb, new_ub);
	}
}


DRangeP RangeMove(DRangeP range, RANGE_VALUE_TYPE amount) {
	return RangeMakeSingle((range->lb + amount), (range->ub + amount));
}

DRangeP RangeShift(DRangeP range) {
	return RangeMakeSingle((range->lb + rangeShiftValue), (range->ub + rangeShiftValue));
}

DRangeP RangeClone(DRangeP orig) {
	if (orig != NULL) {
		return RangeMakeSingle(orig->lb, orig->ub);
	} else {
		return NULL;
	}
}

void	RangePrint(DRangeP r1) {
	if (r1 == NULL) {
		fprintf(stderr, "(NULL)");
	} else if (r1->lb == r1->ub) {
		fprintf(stderr, "%d", r1->lb);
	} else {
		fprintf(stderr, "%d..%d", r1->lb, r1->ub);
	}
}

int	snRangePrint(char *buf, size_t max_chars, DRangeP r1) {
	char tmp[30]; // 30 >= 1(-) + 10 (int) + 2 ("..") + 10 (int) + 1 ('\0')
	int count;
	if (r1->lb == r1->ub) {
		count = sprintf(tmp, "%d", r1->lb);
	} else {
		count = sprintf(tmp, "%d..%d", r1->lb, r1->ub);
	}
	if ( count < max_chars ) {
		return sprintf(buf, "%s", tmp);
	} else {
		return -1;
	}
}	

int	RangeSize(DRangeP r1) {
	return (int) (r1->ub - r1->lb);
}

char	RangeGt(DRangeP r1, RANGE_VALUE_TYPE value) {
	return (r1->ub > value);
}
char	RangeGtEq(DRangeP r1, RANGE_VALUE_TYPE value) {
	return (r1->ub >= value);
}
char	RangeSt(DRangeP r1, RANGE_VALUE_TYPE value) {
	return (r1->lb < value);
}
char	RangeStEq(DRangeP r1, RANGE_VALUE_TYPE value) {
	return (r1->lb <= value);
}

// Private functions

