# =========================================================================
# $Id: tm.tcl,v 1.1.2.6 2000/09/06 16:26:36 belinfan Exp $
# =========================================================================

# tree-matching library.
#
# input:
#  - list of three-tuples < <position, value> list, target >
#    to be used to build the mapping table
#
# output:
#  - a routine that maps items to targets, given
#     - a routine to compare values
#     - a routine to get a value at a position from the item
#     - a routine to get an arity at a position from the item

# interface:
#  tm_init: (:position -> int), (:value -> int), (:target -> int) -> void
#  tm_add_pattern: <type, position, val>-list, target -> { success, failure }
#     (failure may result if the pattern introduces non-confluency)
#     (type = 'n' | 'a'   )
#     (val = name | arity )
#  tm_match: item,
#            (:item, position -> value),
#            (:item, position -> arity) -> target

proc tm_init {poscmp valuecmp targetcmp} {
	global tm
	tm_finalize
	set tm(nrofrules) 0
	set tm(cmp,pos) $poscmp
	set tm(cmp,value) $valuecmp
	set tm(cmp,target) $targetcmp
	set tm(match) tm_interpreted
	return ""
}

proc tm_finalize {} {
	global tm
	catch { unset tm }
	set tm(nrofrules) 0
	return ""
}

proc tm_show {} {
	global tm
	if {![info exists tm(nrofrules)]} {
		puts stderr "tm: inialize first!"
		return -1
	}
	set rulenr 0
	while {$rulenr < $tm(nrofrules)} {
		puts stderr "Rule: $tm(rule,repr,$rulenr) -> $tm(rule,target,$rulenr)"
		foreach p $tm(rule,pairs,$rulenr) {
			puts stderr "\t$p"
		}
		incr rulenr
	}
}

proc tm_list {} {
	global tm
	set rules {}
	if {![info exists tm(nrofrules)]} {
		puts stderr "tm: inialize first!"
		return $rules
	}
	set rulenr 0
	while {$rulenr < $tm(nrofrules)} {
		lappend rules [list $tm(rule,pairs,$rulenr) $tm(rule,target,$rulenr)]
		incr rulenr
	}
	return $rules
}

proc tm_add_pattern { repr pos_value_pairs target } {
	global tm
	if {![info exists tm(nrofrules)]} {
		puts stderr "tm: inialize first!"
		return -1
	}
	set sorted_pairs [tm_sort_pairs $pos_value_pairs]
	return [tm_add_sorted_pattern $repr $sorted_pairs $target]
}
proc tm_add_sorted_pattern { repr pos_value_pairs target } {
	global tm
	if {[tm_remains_confluent $repr $pos_value_pairs $target]} {
		set rulenr $tm(nrofrules)
		set tm(rule,repr,$rulenr) $repr
		set tm(rule,pairs,$rulenr) $pos_value_pairs
		set tm(rule,target,$rulenr) $target
		# debugMsg "tm_add_sorted_pattern ($pos_value_pairs)($target)"
		incr tm(nrofrules)
		set tm(match) tm_interpreted
		return 1
	} else {
		return 0
	}
}

proc tm_sort_pairs {pairs} {
	return [lsort -command tm_paircmp $pairs]
}

proc tm_paircmp {a b} {
	global tm
	set tmp [$tm(cmp,pos) [lindex $a 1] [lindex $b 1]]
	if {$tmp != 0} {
		return $tmp
	} else {
		return [tm_typecmp [lindex $a 0] [lindex $b 0]]
	}
}
proc tm_typecmp {a b} {
	global tm
	if {($a == "a") && ($b == "n")} {
		return -1
	} elseif {($a == "n") && ($b == "a")} {
		return 1
	} else {
		return 0
	}
}

proc tm_remains_confluent { repr pos_value_pairs target } {
	global tm
	set rulenr 0
	while {$rulenr < $tm(nrofrules)} {
		if {[tm_target_cmp $tm(rule,target,$rulenr) $target] != 0} {
			if {![tm_pattern_distinct $tm(rule,pairs,$rulenr) $pos_value_pairs]} {
				puts stderr "Clash: $repr -> $target"
				puts stderr "with:  $tm(rule,repr,$rulenr) -> $tm(rule,target,$rulenr)"
				return 0
			}
		}
		incr rulenr
	}
	return 1
}

proc tm_target_cmp { t1 t2 } {
	global tm
	return [$tm(cmp,target) $t1 $t2]
}

proc tm_pattern_distinct { pat1 pat2 } {
	global tm
	set l1 [llength $pat1]
	set l2 [llength $pat2]
	set i1 0
	set i2 0
	while { ($i1 < $l1) && ($i2 < $l2) } {
		set pos1 [lindex [lindex $pat1 $i1] 1]
		set pos2 [lindex [lindex $pat2 $i2] 1]
		set cmp [$tm(cmp,pos) $pos1 $pos2]
		if {$cmp < 0} {
			incr i1
		} elseif  {$cmp > 0} {
			incr i2
		} else {
			set typ1 [lindex [lindex $pat1 $i1] 0]
			set typ2 [lindex [lindex $pat2 $i2] 0]
			set val1 [lindex [lindex $pat1 $i1] 2]
			set val2 [lindex [lindex $pat2 $i2] 2]
			if {[tm_typecmp $typ1 $typ2] != 0} {
				incr i1
				incr i2
			} elseif {[$tm(cmp,value) $val1 $val2] != 0} {
				# debugMsg "tm_pattern_distinct ($pat1)($pat2): 1"
				return 1
			} else {
				incr i1
				incr i2
			}
		}
	}
	# debugMsg "tm_pattern_distinct ($pat1)($pat2): 0"
	return 0
}

proc tm_match { item getpos getarity } {
	global tm
	set routine tm_interpreted
	return [$tm(match) $item $getpos $getarity]
}

proc tm_interpreted { item getpos getarity } {
	global tm
	if {![info exists tm(nrofrules)]} {
		puts stderr "tm: inialize first!"
		return -1
	}
	tm_search_init
	set cmp $tm(cmp,value)
	set rulenr 0
	while {$rulenr < $tm(nrofrules)} {
		if {[tm_try_rule $item $rulenr $getpos $getarity]} {
			tm_search_cleanup
			return [list 1 $tm(rule,target,$rulenr)]
		}
		incr rulenr
	}
	tm_search_cleanup
	return [list 0]
}

proc tm_search_init {} {
}
proc tm_search_cleanup {} {
	global tmtmp
	catch { unset tmtmp }
}

proc tm_try_rule { item rulenr getpos getarity} {
	global tm
	foreach p $tm(rule,pairs,$rulenr) {
		if {![tm_try_rule_pair $item $p $getpos $getarity]} {
			return 0
		}
	}
	return 1
}

proc tm_try_rule_pair { item pair getpos getarity} {
	global tm
	set type [lindex $pair 0]
	set pos [lindex $pair 1]
	if {[tm_typecmp $type a] == 0} {
		set val [lindex $pair 2]
		set item_val [tm_get arity $item $pos $getarity]
		if {$val == $item_val} {
			set result 1
		} else {
			set result 0
		}
	} else {
		set val [lindex $pair 2]
		set item_val [tm_get value $item $pos $getpos]
		if {[$tm(cmp,value) $val $item_val] == 0} {
			set result 1
		} else {
			set result 0
		}
	}
	# debugMsg "tm_try_rule_pair $item $pair: $item_val -> $result"
	return $result
}

proc tm_get { kind item pos getkind } {
	global tmtmp
	if {![info exists tmtmp($item,$kind,$pos)]} {
		set tmtmp($item,$kind,$pos) [$getkind $item $pos]
	}
	return $tmtmp($item,$kind,$pos)
}

# =========================================================================

# Code to compule rule-pairs
#
proc tm_compile {} {
	global tm
	if {[catch {eval [tm_code]} foo]} {
		warningMsg $foo
		return 0
	} else {
		if {[string compare [info procs tm_compiled] ""] != 0} {
			set tm(match) tm_compiled
		}
		return 1
	}
}
proc tm_code {} {
	set rules [tm_compile_rules [tm_list]]
	set body [tm_generate_code $rules " " " "]
	set s ""
	append s "proc tm_compiled { item getpos getarity } {\n"
	append s $body
	append s "}\n"
	return $s
}
proc tm_compile_rules {rules} {
	if {[llength $rules] <= 0} {
		return $rules
	}
	set sorted [tm_sort_rules $rules]
	set firstrule [lindex $sorted 0]
	set firstpairs [lindex $firstrule 0]
	set firsttarget [lindex $firstrule 1]
	if {[llength $firstpairs] <= 0} {
		return [list t $firsttarget]
	}
	set firstpair [lindex $firstpairs 0]
	set firstpairtype [lindex $firstpair 0]
	set firstpairpos [lindex $firstpair 1]
	
	foreach r $rules {
		set pairs [lindex $r 0]
		set target [lindex $r 1]
		if {[llength $pairs] <= 0} {
			lappend tmcomptmpdefault $r
		} else {
			set pair [lindex $pairs 0]
			if {[tm_paircmp $firstpair $pair] == 0} {
				set val [lindex $pair 2]
				set tailpairs [lrange $pairs 1 end]
				lappend tmcomptmp($val) [list $tailpairs $target]
			} else {
				lappend tmcomptmpdefault $r
			}
		}
	}
	set res {}
	if {[info exists tmcomptmpdefault]} {
		set defpat [tm_compile_rules $tmcomptmpdefault]
		set resdef $tmcomptmpdefault
	} else {
		set defpat [list d ]
		set resdef {}
	}
	foreach k [array names tmcomptmp] {
		set subpat [tm_compile_rules [concat $tmcomptmp($k) $resdef]]
		lappend res [list  $k $subpat]
	}
	return [list $firstpairtype $firstpairpos $res $defpat ]
}

proc tm_generate_code {rule {indent "  "} {inc "  "} } {
	# debugMsg "tm_generate_code: ($rule)"
	if {[llength $rule] <= 0 } {
		return
	}
	set type [lindex $rule 0]
	switch -- $type {
		  t {
			set target [lindex $rule 1]
			return "[set indent]return \[list  1 [list $target ]\]\n"
		} d {
			return "[set indent]return \[list  0 \]\n"
		} a {
			set routine getarity
		} n {
			set routine getpos
		} default {
			debugMsg "tm_generate_code: Unknown type: ($type)"
		}
	}
	set pos [lindex $rule 1]
	set subrules [lindex $rule 2]
	set default [lindex $rule 3]
	set subindent "[set indent][set inc]"
	set subsubindent "[set subindent][set inc]"
	set s ""
	append s "[set indent]switch -- \[\$[set routine] \$item [list $pos]\] {\n"
	foreach r $subrules {
		set val [lindex $r 0]
		set sub [lindex $r 1]
		append s "[set subindent]$val {\n"
		append s [tm_generate_code $sub $subsubindent $inc]
		append s "[set subindent]}\n"
	}
	append s "[set subindent]default {\n"
	append s [tm_generate_code $default $subsubindent $inc]
	append s "[set subindent]}\n"
	append s "[set indent]}\n"
	return $s
}

# =========================================================================

proc tm_sort_rules {rules} {
	return [lsort -command tm_rulecmp $rules]
}

# compare on pairs, not on result!
proc tm_rulecmp {a b} {
	global tm
	set pairs1 [lindex $a 0]
	set pairs2 [lindex $b 0]
	set l1 [llength $pairs1]
	set l2 [llength $pairs2]
	set i1 0
	set i2 0
	while { ($i1 < $l1) && ($i2 < $l2) } {
		set cmp [tm_paircmp [lindex $pairs1 $i1] [lindex $pairs2 $i2]]
		if {$cmp < 0} {
			return -1
		} elseif  {$cmp > 0} {
			return 1
		} else {
			incr i1
			incr i2
		}
	}
	if {$i1 < $l1} {
		return -1
	} elseif  {$i2 < $l2} {
		return 1
	} else {
		return 0
	}
}


# =========================================================================

# Testing routine.
# Note: this routine depends on the em.tcl file for the interfaces!

proc tm_test {} {
	em_init
	em_add_pattern "aap!noot!*" wim
	em_add_pattern "*!zus!*" schapen
	
	tm_do_test interpreted "bla!zus!mies"  [list 1 schapen]
	tm_do_test interpreted "bla!noot!mies" [list 0]
	tm_do_test interpreted "aap!noot!mies" [list 1 wim]
	tm_do_test interpreted "bla!bla!bla"   [list 0]
	
	tm_compile
	
	tm_do_test compiled  "bla!zus!mies"  [list 1 schapen]
	tm_do_test compiled  "bla!noot!mies" [list 0]
	tm_do_test compiled  "aap!noot!mies" [list 1 wim]
	tm_do_test compiled  "bla!bla!bla"   [list 0]

	em_init
	em_add_pattern "aap!noot" wim
	em_add_pattern "*!noot" wim
	
	tm_do_test interpreted "aap!noot"    [list 1 wim]
	tm_do_test interpreted "bla!noot"    [list 1 wim]
	tm_do_test interpreted "bla!bla"     [list 0]
	tm_do_test interpreted "bla!bla!bla" [list 0]
	
	tm_compile
	
	tm_do_test compiled "aap!noot"    [list 1 wim]
	tm_do_test compiled "bla!noot"    [list 1 wim]
	tm_do_test compiled "bla!bla"     [list 0]
	tm_do_test compiled "bla!bla!bla" [list 0]
}

proc tm_do_test {cmt s t} {
	set res_t [em_find_string $s]
	if {[tm_target_cmp $res_t $t] != 0} {
		puts "FAIL $cmt: s ($s) : result ($res_t) : expected ($t)"
	}
}

# =========================================================================
