#include "caesar_graph.h"
#include <stdio.h>
#include <stdarg.h>
#include <unistd.h> /* for pipe() */
#include <string.h> /* SENVA for strdup() */
#include <stdlib.h> /* SENVA for getenv() */

static void Warning(const char *format, ...)
{
  va_list args;
  va_start(args, format);
  fprintf(stderr,"torx_open: warning: ");
  vfprintf(stderr, format, args);
  fprintf(stderr,"\n");
  va_end(args);
}

static void Error(const char *format, ...)
{
  va_list args;
  va_start(args, format);
  fprintf(stderr,"torx_open: ");
  vfprintf(stderr, format, args);
  fprintf(stderr,"\n");
  va_end(args);
  exit(-1);
}

static int debug = 0;
static void Debug(int d, const char *format, ...)
{
  va_list args;
  va_start(args, format);
  if (debug >= d) {
    fprintf(stderr,"torx_open: debug: ");
    vfprintf(stderr, format, args);
    fprintf(stderr,"\n");
  }
  va_end(args);
}

/* ADMINISTRATIVE STUFF */

CAESAR_TYPE_STRING CAESAR_GRAPH_COMPILER() {
  return "torx_open";
}

CAESAR_TYPE_VERSION CAESAR_GRAPH_VERSION() {
  return 2.14;
}

/* ENCODING OF STATES AND LABELS */

struct CAESAR_STRUCT_STATE {int state; char *desc;};
struct CAESAR_STRUCT_LABEL {char* label; int visible;};

CAESAR_TYPE_NATURAL CAESAR_HINT_SIZE_STATE = sizeof(struct CAESAR_STRUCT_STATE);
CAESAR_TYPE_NATURAL CAESAR_HINT_SIZE_LABEL = sizeof(struct CAESAR_STRUCT_LABEL);

CAESAR_TYPE_NATURAL CAESAR_HINT_HASH_SIZE_STATE = sizeof(int);
CAESAR_TYPE_NATURAL CAESAR_HINT_HASH_SIZE_LABEL = sizeof(char*);

CAESAR_TYPE_NATURAL CAESAR_HINT_ALIGNMENT_STATE = 4;
CAESAR_TYPE_NATURAL CAESAR_HINT_ALIGNMENT_LABEL = 4;

CAESAR_TYPE_BOOLEAN CAESAR_COMPARE_STATE(CAESAR_TYPE_STATE s1, CAESAR_TYPE_STATE s2) {
  return (s1->state == s2->state ? CAESAR_TRUE : CAESAR_FALSE);
}

CAESAR_TYPE_BOOLEAN CAESAR_COMPARE_LABEL(CAESAR_TYPE_LABEL l1, CAESAR_TYPE_LABEL l2) {
  return (strcmp(l1->label,l2->label) == 0 ? CAESAR_TRUE : CAESAR_FALSE);
}

/* PRINTING OF STATES AND LABELS */

static state_format = 0;
CAESAR_TYPE_NATURAL CAESAR_MAX_FORMAT_LABEL(void) {return 0;}
CAESAR_TYPE_NATURAL CAESAR_MAX_FORMAT_STATE(void) {return 1;}

void CAESAR_FORMAT_STATE(CAESAR_TYPE_NATURAL x) {
  if (0<=x && x<=1)
    state_format = x;
  else
    Warning("%d exceeds max state format",x);
}

void CAESAR_FORMAT_LABEL(CAESAR_TYPE_NATURAL x) {
  if (x != 0)
    Warning("%d exceeds max label format",x);
}

void CAESAR_PRINT_STATE(CAESAR_TYPE_FILE f,CAESAR_TYPE_STATE s) {
  fprintf(f,"state %d desc %s",s->state, s->desc);
}

void CAESAR_PRINT_LABEL(CAESAR_TYPE_FILE f,CAESAR_TYPE_LABEL l) {
  /* fprintf(stderr,"CAESAR_PRINT_LABEL\n"); */
  if (!l->visible)
    fprintf(f,"i");
  else 
    fprintf(f,"%s",l->label);
}

CAESAR_TYPE_STRING CAESAR_STRING_LABEL(CAESAR_TYPE_LABEL l) {
  // fprintf(stderr,"CAESAR_STRING_LABEL\n");
  char* L;
  static char *a = NULL;
  static int siz  = 0;
  if (!l->visible)
    L="i";
  else {
    L=l->label;
    if (!a || strlen(L)>siz) {
      siz = strlen(L);
      a = realloc(a,  siz+5); 
      }
    sprintf(a,"%s",L);
    return a;
    }
  return L;
  }

void CAESAR_PRINT_STATE_HEADER(CAESAR_TYPE_FILE fp) {
    fprintf(fp,"\n");
}

void CAESAR_DELTA_STATE(CAESAR_TYPE_FILE fp,CAESAR_TYPE_STATE s1,CAESAR_TYPE_STATE s2) {
  fprintf(fp,"Diff: %d - %d",s1->state,s2->state);
}


CAESAR_TYPE_STRING CAESAR_INFORMATION_LABEL(CAESAR_TYPE_LABEL l) {
  return "";
}



/* ON THE FLY EXPLORATION */

static int initial_state = 0;
static FILE *totorx, *fromtorx;
static char *buf = 0;
static int nbuf = 0;
static int bufinc = BUFSIZ;

static char* freadline(FILE *f)
{
  char *p, *e;

  Debug(3,"freadline nbuf=%d", nbuf);
  if (! buf) {
    buf = malloc(bufinc);
    if (! buf)
      Error("out of memory");
    nbuf = bufinc;
  }

  if (!fgets(buf, nbuf, fromtorx))
    return 0;

  p = buf;
  while (! (e = strchr(p, '\n'))) {
    Debug(3,"freadline: incomplete line (nbuf=%d): \"%s\"", nbuf, buf);
    buf = realloc(buf, nbuf+bufinc);
    if (! buf)
      Error("out of memory");
    p = buf + nbuf - 1;
    nbuf = nbuf+bufinc;
    if (!fgets(p, bufinc+1, fromtorx))
      return 0;
  }
  return buf;
}

void CAESAR_ITERATE_STATE(CAESAR_TYPE_STATE s1, 
			  CAESAR_TYPE_LABEL l, 
			  CAESAR_TYPE_STATE s2, 
    void (*callback) 
      (CAESAR_TYPE_STATE, CAESAR_TYPE_LABEL, CAESAR_TYPE_STATE)) {
  int stop = 0;
  int phase = 0;
  char *p, *e, *end, *s;

  fprintf(totorx, "e %d\n", s1->state);
  Debug(2,"sent \"e %d\"", s1->state);
  fflush(totorx);
  while (!stop && (buf = freadline(fromtorx))) {
    Debug(3, "got \"%s\"", buf);
    if (! (e = strchr(buf, '\n'))) {
      Error("incomplete (too long) torx line: \"%s\"", buf);
    }
    *e = '\0';
    end = e;
    p = buf;
    while(isspace((int)*p))
      p++;
    if (strncmp(p, "A_", 2) == 0) {
      fprintf(stderr,"%s\n", buf);
      continue;
    }
    switch(phase) {
    case 0:
      if (strncmp(p, "EB", 2) == 0)
        phase = 1;
      else if (strncmp(p, "E0", 2) == 0)
        Error("error while exploring (e %d): %s", s1->state, p+2);
      else
        Error("error parsing torx line (expected EB or E0): \"%s\"", buf);
      break;
    case 1:
      /* Ee event TAB visible TAB solved TAB label TAB preds TAB freevars TAB identical */
      /* now: Ee */
      if (strncmp(p, "EE", 2) == 0) {
        stop = 1;
        continue;
      }
      if (strncmp(p, "Ee", 2) != 0)
        Error("error parsing torx line (expected Ee): \"%s\"", buf);
      p += 2;
      if (!isspace((int)*p))
        Error("error parsing torx line (expected whitespace after Ee): \"%s\"", buf);
      while(isspace((int)*p))
        p++;
      /* now: event TAB */
      e = strchr(p, '\t');
      if (!e)
        Error("error parsing torx line (no first tab): \"%s\"", buf);
      *e = '\0';
      s = e - 1;
      while(isdigit(*s) && s > p)
        s--;
      if (!isdigit(*s))
        s++;
      s2->state = atoi(s);
      s2->desc = strdup(p);
	  Debug(3, " s2->state = %d (from \"%s\")", s2->state, s);
      *e = '\t';
      p = e + 1;
      /* now: visible TAB */
      e = strchr(p, '\t');
      if (!e)
        Error("error parsing torx line (no second tab): \"%s\"", buf);
      *e = '\0';
      l->visible = atoi(p);
      *e = '\t';
      p = e + 1;
      /* now: solved TAB */
      e = strchr(p, '\t');
      if (!e)
        Error("error parsing torx line (no third tab): \"%s\"", buf);
      p = e + 1;
      /* now: label TAB */
      e = strchr(p, '\t');
      if (!e)
        Error("error parsing torx line (no fourth tab): \"%s\"", buf);
      *e = '\0';
      l->label = strdup(p);
      *e = '\t';
      p = e + 1;
      /* now: preds TAB */
      e = strchr(p, '\t');
      if (!e)
        Error("error parsing torx line (no fifth tab): \"%s\"", buf);
      p = e + 1;
      /* now: freevars TAB */
      e = strchr(p, '\t');
      if (!e)
        Error("error parsing torx line (no sixth tab): \"%s\"\n", buf);
      p = e + 1;
      /* now: identical [TAB | NL] */
      e = strchr(p, '\t');
      if (e)
        *e = '\0';
      else
        e = end;
      Debug(3, "e=%x p=%x", e, p);
      if (p != e) {
        s = e - 1;
        while(isdigit(*s) && s > p)
          s--;
        if (!isdigit(*s))
          s++;
        s2->state = atoi(s);
        free(s2->desc);
        s2->desc = strdup(p);
	    Debug(3, " identical s2->state = %d (from \"%s\")", s2->state, s);
      }
      callback(s1,l,s2);
	  Debug(2, "callback(%d,\"%s\",%d)", s1->state, l->label, s2->state);
      break;
    default:
      Error("internal error: no such phase: %d", phase);
    }
  }
}

void CAESAR_START_STATE(CAESAR_TYPE_STATE s) {
  s->state = initial_state;
  s->desc = strdup("root");
}

/* SOME OTHER REQUIRED FUNCTIONS */

CAESAR_TYPE_BOOLEAN CAESAR_VISIBLE_LABEL(CAESAR_TYPE_LABEL x) {
  return (x->visible ? CAESAR_TRUE : CAESAR_FALSE);
}

CAESAR_TYPE_NATURAL CAESAR_HASH_STATE(CAESAR_TYPE_STATE s,CAESAR_TYPE_NATURAL m) {
  return s->state % m;
}

CAESAR_TYPE_NATURAL CAESAR_HASH_LABEL(CAESAR_TYPE_LABEL l,CAESAR_TYPE_NATURAL m) {
  return (CAESAR_TYPE_NATURAL)strlen(CAESAR_STRING_LABEL(l)) % m;
}


CAESAR_TYPE_STRING CAESAR_GATE_LABEL(CAESAR_TYPE_LABEL l) {
  return strtok(CAESAR_STRING_LABEL(l),"! \t");
}

CAESAR_TYPE_NATURAL CAESAR_CARDINAL_LABEL(CAESAR_TYPE_LABEL l) {
  int i = 1;
  while (strchr(CAESAR_STRING_LABEL(l), '!'))
    i++;
  return i;
}

/* INITIALIZATION */

void CAESAR_INIT_GRAPH(void) {
  char *prog;
  int in[2];
  int out[2];
  int child;
  int stop = 0;
  int phase = 0;
  char *p, *e, *s;

  prog = getenv ("TORX_OPEN_EXPLORER");
  if (prog == NULL)
	Error ("undefined environment variable $TORX_OPEN_EXPLORER");

  Debug(2, "starting: \"%s\"", prog);

  if (pipe(in) < 0) {
    perror("torx_open: pipe in failed");
    exit(-1);
  }
  if (pipe(out) < 0) {
    perror("torx_open: pipe out failed");
    exit(-1);
  }
  if ((child = fork()) < 0) {
    perror("torx_open: fork failed");
    exit(-1);
  }
  if (child > 0) {
    close(0);
    dup(in[1]);
    close(1);
    dup(out[1]);
    close(in[0]);
    close(out[0]);
    Debug(2, "about to exec: \"%s\"", prog);
    execl("/usr/bin/sh", "sh", "-c", prog, (char *)0);
  } else {
    totorx = fdopen(in[0], "w");
    if (! totorx) {
      perror("torx_open: could not fdopen totorx");
      exit(-1);
    }
    fromtorx = fdopen(out[0], "r");
    if (! fromtorx) {
      perror("torx_open: could not fdopen fromtorx");
      exit(-1);
    }
    close(in[1]);
    close(out[1]);
  }

  Debug(3, "started: \"%s\"", prog);

  fprintf(totorx, "r\n");
  fflush(totorx);
  Debug(2, "sent \"r\"");
  while (!stop && (buf = freadline(fromtorx))) {
    Debug(3, "got \"%s\"", buf);
    if (! (e = strchr(buf, '\n'))) {
      Error("incomplete (too long) torx line: \"%s\"", buf);
    }
    *e = '\0';
    p = buf;
    while(isspace((int)*p))
      p++;
    /* R init TAB solved TAB label TAB preds TAB freevars TAB identical */
    if (strncmp(p, "A_", 2) == 0) {
      fprintf(stderr,"%s\n", buf);
      continue;
    }
    if (strncmp(p, "R", 1) != 0)
      Error("error parsing torx line (expected R): \"%s\"", buf);
    p += 1;
    if (!isspace((int)*p))
      Error("error parsing torx line (expected whitespace after R): \"%s\"", buf);
    while(isspace((int)*p))
      p++;
    /* now: event TAB */
    e = strchr(p, '\t');
    if (!e)
      Error("error parsing torx line (no first tab): \"%s\"", buf);
    *e = '\0';
    s = e - 1;
    while(isdigit(*s) && s > p)
      s--;
    if (!isdigit(*s))
      s++;
    initial_state = atoi(s);
    *e = '\t';
    p = e + 1;
    stop = 1;
  }
}
